const { logger } = require('../../../logger');
const versioning = require('../../../versioning');
const { platform } = require('../../../platform');

module.exports = {
  detectVulnerabilityAlerts,
};

async function detectVulnerabilityAlerts(input) {
  if (!(input && input.vulnerabilityAlerts)) {
    return input;
  }
  if (input.vulnerabilityAlerts.enabled === false) {
    logger.debug('Vulnerability alerts are disabled');
    return input;
  }
  const alerts = await platform.getVulnerabilityAlerts();
  if (!alerts.length) {
    logger.debug('No vulnerability alerts found');
    if (input.vulnerabilityAlertsOnly) {
      throw new Error('no-vulnerability-alerts');
    }
    return input;
  }
  const config = { ...input };
  const combinedAlerts = {};
  for (const alert of alerts) {
    try {
      if (alert.dismissReason) {
        continue; // eslint-disable-line no-continue
      }
      if (!alert.securityVulnerability.firstPatchedVersion) {
        logger.info(
          { alert },
          'Vulnerability alert has no firstPatchedVersion - skipping'
        );
        continue; // eslint-disable-line no-continue
      }
      const datasourceMapping = {
        MAVEN: 'maven',
        NPM: 'npm',
        NUGET: 'nuget',
        PIP: 'pypi',
        RUBYGEMS: 'bundler',
      };
      const datasource =
        datasourceMapping[alert.securityVulnerability.package.ecosystem];
      if (!combinedAlerts[datasource]) {
        combinedAlerts[datasource] = {};
      }
      const depName = alert.securityVulnerability.package.name;
      if (!combinedAlerts[datasource][depName]) {
        combinedAlerts[datasource][depName] = {
          advisories: [],
          fileNames: [],
        };
      }
      combinedAlerts[datasource][depName].advisories.push(
        alert.securityAdvisory
      );
      const fileName = alert.vulnerableManifestFilename;
      if (!combinedAlerts[datasource][depName].fileNames.includes(fileName)) {
        combinedAlerts[datasource][depName].fileNames.push(fileName);
      }
      const firstPatchedVersion =
        alert.securityVulnerability.firstPatchedVersion.identifier;
      const versionSchemes = {
        maven: 'maven',
        npm: 'npm',
        nuget: 'semver',
        pip_requirements: 'pep440',
        rubygems: 'ruby',
      };
      const version = versioning.get(versionSchemes[datasource]);
      if (version.isVersion(firstPatchedVersion)) {
        if (combinedAlerts[datasource][depName].firstPatchedVersion) {
          if (
            version.isGreaterThan(
              firstPatchedVersion,
              combinedAlerts[datasource][depName].firstPatchedVersion
            )
          ) {
            combinedAlerts[datasource][
              depName
            ].firstPatchedVersion = firstPatchedVersion;
          }
        } else {
          combinedAlerts[datasource][
            depName
          ].firstPatchedVersion = firstPatchedVersion;
        }
      } else {
        logger.info('Invalid firstPatchedVersion: ' + firstPatchedVersion);
      }
    } catch (err) {
      logger.warn({ err }, 'Error parsing vulnerability alert');
    }
  }
  const alertPackageRules = [];
  for (const [datasource, dependencies] of Object.entries(combinedAlerts)) {
    for (const [depName, val] of Object.entries(dependencies)) {
      let prBodyNotes = [];
      try {
        prBodyNotes = ['### GitHub Vulnerability Alerts'].concat(
          val.advisories.map(advisory => {
            let content = '#### ';
            let heading;
            if (advisory.identifiers.some(id => id.type === 'CVE')) {
              heading = advisory.identifiers
                .filter(id => id.type === 'CVE')
                .map(id => id.value)
                .join(' / ');
            } else {
              heading = advisory.identifiers.map(id => id.value).join(' / ');
            }
            if (advisory.references.length) {
              heading = `[${heading}](${advisory.references[0].url})`;
            }
            content += heading;
            content += '\n\n';
            content += advisory.description;
            return content;
          })
        );
      } catch (err) /* istanbul ignore next */ {
        logger.warn({ err }, 'Error generating vulnerability PR notes');
      }
      let matchCurrentVersion;
      if (datasource === 'maven') {
        matchCurrentVersion = `(,${val.firstPatchedVersion})`;
      } else {
        matchCurrentVersion = `< ${val.firstPatchedVersion}`;
      }
      const allowedVersions =
        datasource === 'pypi'
          ? `==${val.firstPatchedVersion}`
          : val.firstPatchedVersion;
      const matchRule = {
        datasources: [datasource],
        packageNames: [depName],
        matchCurrentVersion,
        allowedVersions,
        prBodyNotes,
        force: {
          ...config.vulnerabilityAlerts,
          vulnerabilityAlert: true,
          branchTopic: `${datasource}-${depName}-vulnerability`,
        },
      };
      alertPackageRules.push(matchRule);
    }
  }
  logger.debug({ alertPackageRules }, 'alert package rules');
  config.packageRules = (config.packageRules || []).concat(alertPackageRules);
  return config;
}
